@@ -1,8 +1,8 @@
 /* dbmdelete.c - Remove the key and its associated data from the database.
-   This is the NDBM unix interface name.  */
+   This is the NDBM UNIX interface name.  */
 
 /* This file is part of GDBM, the GNU data base manager.
-   Copyright (C) 1990-2021 Free Software Foundation, Inc.
+   Copyright (C) 1990-2024 Free Software Foundation, Inc.
 
    GDBM is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
