@@ -1,7 +1,7 @@
 /* gdbmexp.c - Export a GDBM database. */
 
 /* This file is part of GDBM, the GNU data base manager.
-   Copyright (C) 2007-2021 Free Software Foundation, Inc.
+   Copyright (C) 2007-2024 Free Software Foundation, Inc.
 
    GDBM is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -65,13 +65,13 @@ gdbm_export_to_file (GDBM_FILE dbf, FILE *fp)
 	  size = htonl (key.dsize);
 	  if (fwrite (&size, sizeof (size), 1, fp) != 1)
 	    goto write_fail;
-	  if (fwrite (key.dptr, key.dsize, 1, fp) != 1)
+	  if (key.dsize > 0 && fwrite (key.dptr, key.dsize, 1, fp) != 1)
 	    goto write_fail;
 
 	  size = htonl (data.dsize);
 	  if (fwrite (&size, sizeof (size), 1, fp) != 1)
 	    goto write_fail;
-	  if (fwrite (data.dptr, data.dsize, 1, fp) != 1)
+	  if (data.dsize > 0 && fwrite (data.dptr, data.dsize, 1, fp) != 1)
 	    goto write_fail;
  	}
       
